"""Functions for morphology modifications in evaluator."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy as np

logger = logging.getLogger(__name__)
ZERO = 1e-6


def taper_function(distance, strength, taper_scale, terminal_diameter, scale=1.0):
    """Function to model tappered AIS."""
    return strength * np.exp(-distance / taper_scale) + terminal_diameter * scale


def synth_axon(sim=None, icell=None, params=None, scale=1.0):
    """Replace axon with tappered axon initial segment.

    Args:
        sim and icell: neuron related arguments
        params (list): fixed parameter for an emodel, should be length, strenght, taper_scale and
            terminal_diameter
        scale (floag): scale parameter for each cell
    """
    for section in icell.axonal:
        sim.neuron.h.delete_section(sec=section)

    sim.neuron.h.execute("create axon[2]", icell)

    nseg_total = 10
    L_target = params[0]
    diameters = taper_function(np.linspace(0, L_target, nseg_total), *params[1:], scale=scale)
    count = 0
    for section in icell.axon:
        section.nseg = nseg_total // 2
        section.L = L_target / 2
        for seg in section:
            seg.diam = diameters[count]
            count += 1

        icell.axonal.append(sec=section)
        icell.all.append(sec=section)

    icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
    icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

    sim.neuron.h.execute("create myelin[1]", icell)
    icell.myelinated.append(sec=icell.myelin[0])
    icell.all.append(sec=icell.myelin[0])
    icell.myelin[0].nseg = 5
    icell.myelin[0].L = 1000
    icell.myelin[0].diam = diameters[-1]  # this assigns the value of terminal_diameter
    icell.myelin[0].connect(icell.axon[1], 1.0, 0.0)


def get_synth_axon_hoc(params):
    # pylint: disable=consider-using-f-string
    return """
proc replace_axon(){ local count, i1, i2, L_target, strenght, taper_scale, terminal_diameter localobj diams

    access axon[0]
    axon[0] i1 = v(0.0001) // used when serializing sections prior to sim start
    axon[1] i2 = v(0.0001) // used when serializing sections prior to sim start
    axon[2] i3 = v(0.0001) // used when serializing sections prior to sim start

    // get rid of the old axon
    forsec axonal{delete_section()}
    execute1("create axon[2]", CellRef)

    // creating diameter profile
    nseg_total = 10
    L_target = %s
    strength = %s
    taper_scale = %s
    terminal_diameter = %s
    scale = $1
    diams = new Vector()
    count = 0
    for i=0,nseg_total{
        count = count + 1
        diams.resize(count)
        diams.x[count-1] = strength * exp(-L_target * i / nseg_total / taper_scale) + terminal_diameter * scale
    }

    // assigning diameter to axon
    count = 0
    for i=0,1{
        access axon[i]
        L =  L_target/2
        nseg = nseg_total/2
        for (x) {
            if (x > 0 && x < 1) {
                diam(x) = diams.x[count]
                count = count + 1
            }
        }
        all.append()
        axonal.append()

        if (i == 0) {
            v(0.0001) = i1
        } else {
            v(0.0001) = i2
        }


    }

    soma[0] connect axon[0](0), 1
    axon[0] connect axon[1](0), 1

    // add myelin part
    create myelin[1]
    access myelin{
            L = 1000
            diam = diams.x[count-1]
            nseg = 5
            v(0.0001) = i3
            all.append()
            myelinated.append()
    }
    connect myelin(0), axon[1](1)
}
    """ % tuple(
        params
    )


def replace_axon_with_taper(sim=None, icell=None):
    """Replace axon with tappered axon initial segment"""
    if len(list(icell.axonal)) < 3:
        raise ValueError(
            "Less than three axon sections are present! "
            "The replace_axon_with_taper morph modifier cannot be applied to such a morphology. "
            "Please use 'bluepyopt_replace_axon' or other valid modifiers from "
            "https://github.com/BlueBrain/BluePyEModel/blob/main/"
            "bluepyemodel/evaluation/modifiers.py or a custom modifier."
        )

    L_target = 60  # length of stub axon
    nseg0 = 5  # number of segments for each of the two axon sections

    nseg_total = nseg0 * 2
    chunkSize = L_target / nseg_total

    diams = []
    lens = []

    count = 0
    for section in icell.axonal:
        L = section.L
        nseg = 1 + int(L / chunkSize / 2.0) * 2  # nseg to get diameter
        section.nseg = nseg

        for seg in section:
            count = count + 1
            diams.append(seg.diam)
            lens.append(L / nseg)
            if count == nseg_total:
                break
        if count == nseg_total:
            break

    for section in icell.axonal:
        sim.neuron.h.delete_section(sec=section)

    #  new axon array
    sim.neuron.h.execute("create axon[2]", icell)

    L_real = 0
    count = 0
    for _, section in enumerate(icell.axon):
        section.nseg = nseg_total // 2
        section.L = L_target / 2

        for seg in section:
            if count >= len(diams):
                break
            seg.diam = diams[count]
            L_real = L_real + lens[count]
            count = count + 1

        icell.axonal.append(sec=section)
        icell.all.append(sec=section)

        if count >= len(diams):
            break
    # childsec.connect(parentsec, parentx, childx)
    icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
    icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

    sim.neuron.h.execute("create myelin[1]", icell)
    icell.myelinated.append(sec=icell.myelin[0])
    icell.all.append(sec=icell.myelin[0])
    icell.myelin[0].nseg = 5
    icell.myelin[0].L = 1000
    icell.myelin[0].diam = diams[count - 1]
    icell.myelin[0].connect(icell.axon[1], 1.0, 0.0)

    logger.debug(
        "Replace axon with tapered AIS of length %f, target length was %f, diameters are %s",
        L_real,
        L_target,
        diams,
    )


def remove_soma(sim=None, icell=None):
    """Remove the soma and connect dendrites together.

    For this to work, we leave the soma connected to the axon,
    and with diameter 1e-6. BluePyOp requires a soma for
    parameter scaling, and NEURON fails is the soma size is =0.
    """
    for section in icell.basal:
        if section.parentseg().sec in list(icell.soma):
            sim.neuron.h.disconnect(section)
            section.connect(icell.axon[0])

    for section in icell.apical:
        if section.parentseg().sec in list(icell.soma):
            sim.neuron.h.disconnect(section)
            section.connect(icell.axon[0])

    for section in icell.soma:
        section.diam = ZERO

    logger.debug("Remove soma")


def isolate_soma(sim=None, icell=None):
    """Remove everything except the soma."""
    for section in icell.axonal:
        sim.neuron.h.delete_section(sec=section)
    for section in icell.basal:
        sim.neuron.h.delete_section(sec=section)
    for section in icell.apical:
        sim.neuron.h.delete_section(sec=section)

    logger.debug("Keep only soma")


def remove_axon(sim=None, icell=None):  # pylint: disable=unused-argument
    """Remove the axon.

    Removing the axonal branches breaks the code, so we set it to ZERO
    """
    for section in icell.myelin:
        sim.neuron.h.delete_section(sec=section)
    for section in icell.axonal:
        section.diam = ZERO


def isolate_axon(sim=None, icell=None):
    """Remove everything except the axon."""
    for section in icell.basal:
        sim.neuron.h.delete_section(sec=section)
    for section in icell.apical:
        sim.neuron.h.delete_section(sec=section)
    for section in icell.soma:
        sim.neuron.h.delete_section(sec=section)


replace_axon_hoc = """
    proc replace_axon(){ local nSec, L_chunk, dist, i1, i2, count, L_target, chunkSize, L_real localobj diams, lens

        L_target = 60  // length of stub axon
        nseg0 = 5  // number of segments for each of the two axon sections

        nseg_total = nseg0 * 2
        chunkSize = L_target/nseg_total

        nSec = 0
        forsec axonal{nSec = nSec + 1}

        // Try to grab info from original axon
        if(nSec < 3){ //At least two axon sections have to be present!

            execerror("Less than three axon sections are present! This emodel can't be run with such a morphology!")

        } else {

            diams = new Vector()
            lens = new Vector()

            access axon[0]
            axon[0] i1 = v(0.0001) // used when serializing sections prior to sim start
            axon[1] i2 = v(0.0001) // used when serializing sections prior to sim start
            axon[2] i3 = v(0.0001) // used when serializing sections prior to sim start

            count = 0
            forsec axonal{ // loop through all axon sections

                nseg = 1 + int(L/chunkSize/2.)*2  //nseg to get diameter

                for (x) {
                    if (x > 0 && x < 1) {
                        count = count + 1
                        diams.resize(count)
                        diams.x[count-1] = diam(x)
                        lens.resize(count)
                        lens.x[count-1] = L/nseg
                        if( count == nseg_total ){
                            break
                        }
                    }
                }
                if( count == nseg_total ){
                    break
                }
            }

            // get rid of the old axon
            forsec axonal{delete_section()}
            execute1("create axon[2]", CellRef)

            L_real = 0
            count = 0

            // new axon dependant on old diameters
            for i=0,1{
                access axon[i]
                L =  L_target/2
                nseg = nseg_total/2

                for (x) {
                    if (x > 0 && x < 1) {
                        diam(x) = diams.x[count]
                        L_real = L_real+lens.x[count]
                        count = count + 1
                    }
                }

                all.append()
                axonal.append()

                if (i == 0) {
                    v(0.0001) = i1
                } else {
                    v(0.0001) = i2
                }
            }

            nSecAxonal = 2
            soma[0] connect axon[0](0), 1
            axon[0] connect axon[1](0), 1

            create myelin[1]
            access myelin{
                    L = 1000
                    diam = diams.x[count-1]
                    nseg = 5
                    v(0.0001) = i3
                    all.append()
                    myelinated.append()
            }
            connect myelin(0), axon[1](1)
        }
    }
"""


def replace_axon_legacy(sim=None, icell=None):
    """Replace axon used in legacy thalamus project"""
    if len(list(icell.axonal)) < 2:
        raise ValueError(
            "Less than two axon sections are present! "
            "The replace_axon_legacy morph modifier cannot be applied to such a morphology. "
            "Please use 'bluepyopt_replace_axon' or other valid modifiers from "
            "https://github.com/BlueBrain/BluePyEModel/blob/main/"
            "bluepyemodel/evaluation/modifiers.py or a custom modifier."
        )

    L_target = 60  # length of stub axon
    nseg0 = 5  # number of segments for each of the two axon sections

    nseg_total = nseg0 * 2
    chunkSize = L_target / nseg_total

    diams = []
    lens = []

    count = 0
    for section in icell.axonal:
        L = section.L
        nseg = 1 + int(L / chunkSize / 2.0) * 2  # nseg to get diameter
        section.nseg = nseg

        for seg in section:
            diams.append(seg.diam)
            lens.append(L / nseg)
            if count == nseg_total:
                break
        if count == nseg_total:
            break

    for section in icell.axonal:
        sim.neuron.h.delete_section(sec=section)

    #  new axon array
    sim.neuron.h.execute("create axon[2]", icell)

    L_real = 0
    count = 0

    for section in icell.axon:
        section.nseg = nseg_total // 2
        section.L = L_target / 2

        for seg in section:
            seg.diam = diams[count]
            L_real = L_real + lens[count]
            count = count + 1
            # print seg.x, seg.diam
        icell.axonal.append(sec=section)
        icell.all.append(sec=section)

    icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
    icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

    logger.debug("Replace axon with tapered AIS")


replace_axon_legacy_hoc = """
proc replace_axon(){local nSec, L_chunk, dist, i1, i2, count, L_target, chunkSize, L_real localobj   diams, lens

    L_target = 60  // length of stub axon
    nseg0 = 5  // number of segments for each of the two axon sections

    nseg_total = nseg0 * 2
    chunkSize = L_target/nseg_total

    nSec = 0
    forsec axonal{nSec = nSec + 1}

    // Try to grab info from original axon
    //At least two axon sections have to be present!
    if(nSec < 2){
        execerror("Less than two axon sections are present! Add an axon to the morphology and try again!")
    } else {

        diams = new Vector()
        lens = new Vector()

        access axon[0]
        i1 = v(0.0001) // used when serializing sections prior to sim start
        access axon[1]
        i2 = v(0.0001) // used when serializing sections prior to sim start

        count = 0
        forsec axonal{ // loop through all axon sections

            nseg = 1 + int(L/chunkSize/2.)*2  //nseg to get diameter

            for (x) {
                if (x > 0 && x < 1) {
                    count = count + 1
                    diams.resize(count)
                    diams.x[count-1] = diam(x)
                    lens.resize(count)
                    lens.x[count-1] = L/nseg
                    if( count == nseg_total ){
                        break
                    }
                }
            }
            if( count == nseg_total ){
                break
            }
        }

        // get rid of the old axon
        forsec axonal{delete_section()}
        execute1("create axon[2]", CellRef)

        L_real = 0
        count = 0

        // new axon dependant on old diameters
        for i=0,1{

            access axon[i]
            L =  L_target/2
            nseg = nseg_total/2

            for (x) {
                if (x > 0 && x < 1) {
                    diam(x) = diams.x[count]
                    L_real = L_real+lens.x[count]
                    count = count + 1
                }
            }

            all.append()
            axonal.append()

            if (i == 0) {
                v(0.0001) = i1
            } else {
                v(0.0001) = i2
            }
        }

        nSecAxonal = 2
        soma[0] connect axon[0](0), 1
        axon[0] connect axon[1](0), 1

        //print 'Target stub axon length:', L_target, 'um, equivalent length: ', L_real 'um'
    }
}
"""


def replace_axon_olfactory_bulb(sim=None, icell=None):
    """Replace axon used in olfactory bulb models."""
    for section in icell.axonal:
        sim.neuron.h.delete_section(sec=section)

    sim.neuron.h.execute("create hillock", icell)
    sim.neuron.h.execute("create initialseg", icell)
    sim.neuron.h.execute("create node[5]", icell)
    sim.neuron.h.execute("create myelin[5]", icell)

    section = icell.hillock
    section.nseg = 3
    section.L = 5
    for seg in section:
        seg.diam = 10.0925
    icell.all.append(sec=section)
    icell.axonal.append(sec=section)

    section = icell.initialseg
    section.nseg = 3
    section.L = 30
    for seg in section:
        seg.diam = 1.5
    icell.all.append(sec=section)
    icell.axonal.append(sec=section)

    for _, section in enumerate(icell.node):
        section.L = 1
        for seg in section:
            seg.diam = 1

        icell.all.append(sec=section)
        icell.axonal.append(sec=section)

    for _, section in enumerate(icell.myelin):
        section.L = 1000
        for seg in section:
            seg.diam = 1.5

        icell.all.append(sec=section)
        icell.myelinated.append(sec=section)

    icell.hillock.connect(icell.soma[0], 0.0, 0.0)
    icell.initialseg.connect(icell.hillock, 1.0, 0.0)

    icell.myelin[0].connect(icell.initialseg, 1.0, 0.0)
    icell.node[0].connect(icell.myelin[0], 1.0, 0.0)
    icell.myelin[1].connect(icell.node[0], 1.0, 0.0)
    icell.node[1].connect(icell.myelin[1], 1.0, 0.0)
    icell.myelin[2].connect(icell.node[1], 1.0, 0.0)
    icell.node[2].connect(icell.myelin[2], 1.0, 0.0)
    icell.myelin[3].connect(icell.node[2], 1.0, 0.0)
    icell.node[3].connect(icell.myelin[3], 1.0, 0.0)
    icell.myelin[4].connect(icell.node[3], 1.0, 0.0)
    icell.node[4].connect(icell.myelin[4], 1.0, 0.0)


replace_axon_olfactory_bulb_hoc = """
proc replace_axon(){ local nSec, L_chunk, dist, i1, i2, count, L_target, chunkSize, L_real localobj diams, lens

    L_target = 60  // length of stub axon
    nseg0 = 5  // number of segments for each of the two axon sections

    nseg_total = nseg0 * 2
    chunkSize = L_target/nseg_total

    nSec = 0
    forsec axonal{nSec = nSec + 1}

    // Try to grab info from original axon
    if(nSec < 1){ //At least two axon sections have to be present!

        execerror("Less than two axon sections are present! Add an axon to the morphology and try again!")

    } else {

        diams = new Vector()
        lens = new Vector()

        access axon[0]
        i1 = v(0.0001) // used when serializing sections prior to sim start

        access axon[1]
        i2 = v(0.0001) // used when serializing sections prior to sim start

        count = 0

        forsec axonal{ // loop through all axon sections

            nseg = 1 + int(L/chunkSize/2.)*2  //nseg to get diameter

        for (x) {
            if (x > 0 && x < 1) {
                count = count + 1
                diams.resize(count)
                diams.x[count-1] = diam(x)
                lens.resize(count)
                lens.x[count-1] = L/nseg
                if( count == nseg_total ){
                    break
                }
            }
        }
        if( count == nseg_total ){
            break
        }
    }

    forsec axonal{delete_section()}
    execute1("create hillock", CellRef)
    execute1("create initialseg", CellRef)
    execute1("create node[5]", CellRef)
    execute1("create myelin[5]", CellRef)


    access hillock
    nseg = 3
    L=5
    diam = 10.0925
    v(0.0001) = i1
    nseg=3
    all.append
    axonal.append
    access initialseg
    nseg = 3
    L=30
    diam=1.5
    v(0.0001) = i2
    nseg=3
    all.append
    axonal.append

    index = i2
    for i=0,4{
        access node[i]
        index += 1
        L=1
        diam=1
        v(0.0001) = index
        all.append()
        axonal.append()
    }
    for i=0,4{
        access myelin[i]
        index += 1
        v(0.0001) = index
        L=1000
        diam=1.5
        v(0.0001) = i2
        all.append()
        myelinated.append()
    }

    nSecAxonal = 12
    connect hillock(0), soma(0)
    connect initialseg(0), hillock(1)

    connect myelin[0](0), initialseg(1)
    connect node[0](0), myelin[0](1)
    connect myelin[1](0), node[0](1)
    connect node[1](0), myelin[1](1)
    connect myelin[2](0), node[1](1)
    connect node[2](0), myelin[2](1)
    connect myelin[3](0), node[2](1)
    connect node[3](0), myelin[3](1)
    connect myelin[4](0), node[3](1)
    connect node[4](0), myelin[4](1)
    }
}
"""



""" ICSELECTOR
Select a set of NEURON mechanisms, parameters and bounds from corresponding
genes. Gene names can be selected from a file mapping genes to different
ME-T types. Corresponding channels are selected from a file mapping channels
and parameters to genes.

## Usage
Select from a gene-mapping file:
    $ python icselector.py
        --ic_map <ic_mapping_file.json>
        --gene_map <gene_mapping_file.csv>
        --keys <any_key_from_gene_map> ...

<any_key_from_gene_map> could be an me-type, t-type or gene name from the
gene_map file, or part of a name e.g.
        --keys L3_TPC:A 'L2/3 IT Cxcl14_1'
"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from argparse import ArgumentParser
from pprint import pprint

from bluepyemodel.icselector import ICSelector


# Get command line arguments
parser = ArgumentParser(
    description="Retrieve a list of NEURON mechanisms associated with provided genes.",
    usage="%(prog)s",
)
parser.add_argument(
    "--ic_map",
    dest="ic_map_path",
    type=str,
    help="Path to .json file containing gene to channel mapping.",
)
parser.add_argument(
    "--gene_map",
    dest="gene_map_path",
    type=str,
    help="Path to .csv file containing met-type to gene mapping.",
)
parser.add_argument(
    "--keys",
    type=str,
    nargs="+",
    help="Optional list of keywords to filter genes.",
)
parser.add_argument(
    "--mode",
    type=str,
    choices=["genetic", "generic", "mixed"],
    default="mixed",
    help="Model types to include.",
)
parser.add_argument(
    "--status",
    type=str,
    choices=["stable", "latest"],
    default="latest",
    help="Model version to include.",
)
args = parser.parse_args()
cmd_args = vars(args)

# === Instantiate ICSelector
keys = cmd_args.pop("keys", None)
icselector = ICSelector(**cmd_args)

# === Get cell configuration
parameters, mechanisms, distributions, nexus_keys = icselector.get_cell_config_from_ttype(keys)
pprint(parameters)
pprint(mechanisms)
pprint(distributions)
pprint(nexus_keys)

# === Retrieve all mechanisms available in Nexus
mechs = icselector.get_mechanisms(selected_only=True)
pprint(mechs)

# === Retrieve channels mapped from genes
genes = icselector.get_gene_mapping()
pprint(genes)



"""Data utils"""

"""
Copyright 2023-2024, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import csv
from importlib import resources

import numpy as np


def get_dendritic_data_filepath(data_type):
    """Gets a dendritic data file path.

    Args:
        data_type (str): can be 'ISI_CV' or 'rheobase'

    Raises:
        ValueError if data_type is not 'ISI_CV' nor 'rheobase'
    """
    if data_type == "ISI_CV":
        return resources.files("bluepyemodel").joinpath("data", "ISI_CV_Shai2015.csv")
    if data_type == "rheobase":
        return resources.files("bluepyemodel").joinpath(
            "data", "spike_rheobase_pA_BeaulieuLaroche2021.csv"
        )

    raise ValueError(f"read_data expects 'ISI_CV' or 'rheobase' but got {data_type}")


def read_dendritic_data(data_type):
    """Reads a dendritic data file and returns distance and data values.

    rheobase values are returned in nA.

    Args:
        data_type (str): can be 'ISI_CV' or 'rheobase'
    """
    file_path = get_dendritic_data_filepath(data_type)

    data = []
    distances = []
    with open(file_path, "r") as csvfile:
        csvreader = csv.reader(csvfile)
        for distance, dat in csvreader:
            distances.append(float(distance))
            data.append(float(dat))

    if data_type == "rheobase":
        data = np.asarray(data) / 1000.0  # pA -> nA
    return np.asarray(distances), data



"""Mechanism class corresponding to mechanisms fields in the icmapping file."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from dataclasses import asdict
from dataclasses import dataclass


@dataclass
class Distribution:
    """Holds subcellular distribution for each compartment."""

    all: str = ""  # If 'all' is set, other fields will be ignored!
    somatic: str = "uniform"
    basal: str = "uniform"
    apical: str = "uniform"
    axonal: str = "uniform"
    myelinated: str = ""

    def set_all(self, dist):
        """Set distribution on all compartments, but not on the 'all' field.

        Args:
            dist (str): key of the subcellular distribution
        """

        fields = asdict(self)
        fields.pop("all")
        for field in fields:
            setattr(self, field, dist)

    def set_fields(self, **fields):
        """Set distributions on compartments.

        Args:
            fields (dict): fields to be set specified as
                {compartment: distribution}
        """

        # If 'fields' contains 'all', then 'all' is set and other compartments
        # are emptied
        if "all" in fields:
            if not fields["all"] == "":
                self.set_all("")
                self.all = fields["all"]
                return

        for k, v in fields.items():
            if isinstance(v, str):
                if v == "nan":
                    v = ""
                setattr(self, k, v)
            else:
                setattr(self, k, "")

    def get(self):
        """Get distributions as a dict.

        Returns:
            distr (dict): distributions for all comparments"""

        distr = {k: v for k, v in asdict(self).items() if not v == ""}
        return distr

    def __str__(self):
        out_str = []
        distr = self.get()
        for k, v in distr.items():
            out_str += [f"{k} = {v}"]
        return ", ".join(out_str)



"""SpikeRec stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class SpikeRecMultiSpikes(BPEM_stimulus):
    # pylint: disable=line-too-long

    """SpikeRecMultiSpikes current stimulus

    .. code-block:: none

              holdi        holdi+amp        holdi       holdi+amp            .   .   .
                :               :             :             :
                :       _________________     :      _________________                      _________________
                :      |                 |    :     |                 |                    |                 |
                :      |                 |    :     |                 |     * n_spikes     |                 |
                :      |                 |    :     |                 |     .   .   .      |                 |
                :      |                 |    :     |                 |                    |                 |
        |______________|                 |__________|                 |__                __|                 |___
        :              :                 :          :                 :                                          ^
        :              :                 :          :                 :                                          :
        :              :                 :          :                 :                                          :
         <--  delay  --><-spike_duration-><- delta -><-spike_duration->     .   .   .                  totduration

    """

    name = "SpikeRecMultiSpikes"

    def __init__(self, location, **kwargs):
        """Constructor

        Attention! This is the class for the new SpikeRec containing multispikes.
        In order to use it, you should use ``SpikeRecMultiSpikes`` as key in your protocol file.
        If you use ``SpikeRec``, it will use the old SpikeRec containing one spike,
        which is using the IDRest class.
        Beware that the `**kwargs` for the two types (multispikes/1spike) of SpikeRec are different.

        Args:
            location(Location): location of stimulus
            **kwargs: See below

        Keyword Arguments:
            amp (float): amplitude of each spike(nA)
            thresh_perc (float): amplitude of each spike relative
                to the threshold current (%)
            holding_current (float): amplitude of the holding current (nA)
            delay (float): time at which the first current spike begins (ms)
            n_spikes (int): number of spikes for the stimulus
            spike_duration (float): duration of each spike (ms)
            delta (float): time without stimulus between each spike (ms)
            totduration (float): total duration of the whole stimulus (ms)
        """

        self.amp = kwargs.get("amp", None)
        self.amp_rel = kwargs.get("thresh_perc", None)

        self.holding_current = kwargs.get("holding_current", None)
        self.threshold_current = None

        if self.amp is None and self.amp_rel is None:
            raise TypeError(f"In stimulus {self.name}, amp and thresh_perc cannot be both None.")

        self.delay = kwargs.get("delay", 10.0)
        self.n_spikes = kwargs.get("n_spikes", 2)
        self.spike_duration = kwargs.get("spike_duration", 3.5)
        self.delta = kwargs.get("delta", 3.5)
        self.total_duration = kwargs.get("totduration", 1500.0)

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.delay

    @property
    def stim_end(self):
        return self.delay + self.n_spikes * self.spike_duration + (self.n_spikes - 1) * self.delta

    @property
    def amplitude(self):
        if self.amp_rel is None or self.threshold_current is None:
            return self.amp
        return self.threshold_current * (float(self.amp_rel) / 100.0)

    def multi_stim_start(self):
        return [self.delay + i * (self.spike_duration + self.delta) for i in range(self.n_spikes)]

    def multi_stim_end(self):
        return [ss + self.spike_duration for ss in self.multi_stim_start()]

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        holding_current = self.holding_current if self.holding_current is not None else 0

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        self.time_vec.append(0.0)
        self.current_vec.append(holding_current)

        spike_start = self.delay
        spike_end = spike_start + self.spike_duration

        self.time_vec.append(spike_start)
        self.current_vec.append(holding_current)

        self.time_vec.append(spike_start)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(spike_end)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(spike_end)
        self.current_vec.append(holding_current)

        for _ in range(1, self.n_spikes):
            spike_start = spike_end + self.delta
            spike_end = spike_start + self.spike_duration

            self.time_vec.append(spike_start)
            self.current_vec.append(holding_current)

            self.time_vec.append(spike_start)
            self.current_vec.append(holding_current + self.amplitude)

            self.time_vec.append(spike_end)
            self.current_vec.append(holding_current + self.amplitude)

            self.time_vec.append(spike_end)
            self.current_vec.append(holding_current)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(holding_current)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def generate(self, dt=0.1):
        """Return current time series"""
        holding_current = self.holding_current if self.holding_current is not None else 0

        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, holding_current, dtype="float64")

        spike_start_idx = int(self.delay / dt)
        spike_end_idx = int((self.delay + self.spike_duration) / dt)
        current[spike_start_idx:spike_end_idx] += self.amplitude

        for _ in range(1, self.n_spikes):
            spike_start_idx = int(spike_end_idx + (self.delta / dt))
            spike_end_idx = spike_start_idx + int(self.spike_duration / dt)
            current[spike_start_idx:spike_end_idx] += self.amplitude

        return t, current



"""Module with recording classes and functions."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy
from bluepyopt import ephys

logger = logging.getLogger(__name__)


def get_loc_ions(isection):
    """Get all ion concentrations available in a location."""
    local_overall_ions = set()

    # ion overall current & concentration
    ions = isection.psection()["ions"]
    for _, ion in ions.items():
        for var in ion.keys():
            # concentration should have 'i' at the end (e.g. ki, cai, nai, ...)
            if var[-1] == "i":
                local_overall_ions.add(var)

    return local_overall_ions


def get_loc_currents(isection):
    """Get all overall currents available in a location."""
    local_overall_currs = set()

    # ion overall current & concentration
    ions = isection.psection()["ions"]
    for _, ion in ions.items():
        for var in ion.keys():
            # current should have 'i' at the beginning (e.g. ik, ica, ina, ...)
            if var[0] == "i":
                local_overall_currs.add(var)

    return local_overall_currs


def get_loc_varlist(isection):
    """Get all possible variables in a location."""
    local_varlist = []

    # currents etc.
    raw_dict = isection.psection()["density_mechs"]
    for channel, vars_ in raw_dict.items():
        for var in vars_.keys():
            local_varlist.append("_".join((var, channel)))

    local_varlist.append("v")

    return local_varlist


def get_i_membrane(isection):
    """Look for i_membrane in a location."""
    raw_dict = isection.psection()["density_mechs"]
    if "extracellular" in raw_dict:
        if "i_membrane" in raw_dict["extracellular"]:
            return ["i_membrane"]
    return []


def check_recordings(recordings, icell, sim):
    """Returns a list of valid recordings (where the variable is in the location)."""

    new_recs = []  # to return
    varlists = {}  # keep varlists to avoid re-computing them every time

    for rec in recordings:
        # get section from location
        try:
            seg = rec.location.instantiate(sim=sim, icell=icell)
        except ephys.locations.EPhysLocInstantiateException:
            continue
        sec = seg.sec
        section_key = str(sec)

        # get list of variables available in the section
        if section_key in varlists:
            local_varlist = varlists[section_key]
        else:
            local_varlist = (
                get_loc_varlist(sec)
                + list(get_loc_ions(sec))
                + list(get_loc_currents(sec))
                + get_i_membrane(sec)
            )
            varlists[section_key] = local_varlist

        # keep recording if its variable is available in its location
        if rec.variable in local_varlist:
            rec.checked = True
            new_recs.append(rec)

    return new_recs


class LooseDtRecordingCustom(ephys.recordings.CompRecording):
    """Recording that can be checked, but that do not records at fixed dt."""

    def __init__(self, name=None, location=None, variable="v"):
        """Constructor.

        Args:
            name (str): name of this object
            location (Location): location in the model of the recording
            variable (str): which variable to record from (e.g. 'v')
        """
        super().__init__(name=name, location=location, variable=variable)

        # important to turn current densities into currents
        self.segment_area = None
        # important to detect ion concentration variable
        self.local_ion_list = None
        self.checked = False

    def instantiate(self, sim=None, icell=None):
        """Instantiate recording."""
        logger.debug("Adding compartment recording of %s at %s", self.variable, self.location)

        self.varvector = sim.neuron.h.Vector()
        seg = self.location.instantiate(sim=sim, icell=icell)
        self.varvector.record(getattr(seg, f"_ref_{self.variable}"))

        self.segment_area = seg.area()
        self.local_ion_list = get_loc_ions(seg.sec)

        self.tvector = sim.neuron.h.Vector()
        self.tvector.record(sim.neuron.h._ref_t)  # pylint: disable=W0212

        self.instantiated = True

    @property
    def response(self):
        """Return recording response. Turn current densities into currents."""

        if not self.instantiated:
            return None

        # do not modify voltage or ion concentration
        if self.variable == "v" or self.variable in self.local_ion_list:
            return ephys.responses.TimeVoltageResponse(
                self.name, self.tvector.to_python(), self.varvector.to_python()
            )

        # ionic current: turn mA/cm2 (*um2) into pA
        return ephys.responses.TimeVoltageResponse(
            self.name,
            self.tvector.to_python(),
            numpy.array(self.varvector.to_python()) * self.segment_area * 10.0,
        )


class FixedDtRecordingCustom(LooseDtRecordingCustom):
    """Recording that can be checked, with recording every 0.1 ms."""

    def instantiate(self, sim=None, icell=None):
        """Instantiate recording."""
        logger.debug("Adding compartment recording of %s at %s", self.variable, self.location)

        self.varvector = sim.neuron.h.Vector()
        seg = self.location.instantiate(sim=sim, icell=icell)
        self.varvector.record(getattr(seg, f"_ref_{self.variable}"), 0.1)

        self.segment_area = seg.area()
        self.local_ion_list = get_loc_ions(seg.sec)

        self.tvector = sim.neuron.h.Vector()
        self.tvector.record(sim.neuron.h._ref_t, 0.1)  # pylint: disable=W0212

        self.instantiated = True
